#!/usr/bin/env python
# NDS texture compressor.
# Copyright (C) 2009 Kevin Vance <kvance@kvance.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""NDS texture compressor."""


import collections
import getopt
import Queue
import struct
import sys
import warnings
import zlib

import Image
import numpy

from scipy.cluster import vq

# Python 2.6+ has pyprocessing in the stdlib, but some of the symbols have
# changed.
try:
    import multiprocessing
except ImportError:
    import processing as multiprocessing
if not hasattr(multiprocessing, 'cpu_count'):
    if not hasattr(multiprocessing, 'cpuCount'):
        raise ImortError, "multiprocessing module missing cpu_count"
    else:
        multiprocessing.cpu_count = multiprocessing.cpuCount


#
# Constants
#

# Python has no infinity constant?
try:
    INF = float('inf')
except ValueError:
    INF = 1e66666 # WINDOWS!

# Program version
VERSION = 0.1

# The maximum sum-of-squared-differences that a 2-color palette can have before
# a 4-color palette is tried.
# XXX this just makes the end result worse.
ERROR_THRESHOLD = 0.0

# The default palette size.
PALETTE_SIZE = 64 * 1024

# The default block modes to try.
# XXX Only allowing mode 2 seems to have the best results.
BLOCK_MODES = [2]

# Blocks with this much error should be redone.
EGREGIOUS_THRESHOLD = 0.3

# Egregiously bad blocks will be acceptible if they have this much error.
CORRECTION_THRESHOLD = 0.1


#
# Block structure
#

class Block:
    """A standalone 4x4 compressed block."""
    def __init__(self, pixmap=0, palette=[], mode=0):
        if type(pixmap) is int:
            self.pixmap = pixmap
        else:
            self.pixmap = 0
            self.copy_pixels(pixmap)
        self.palette = palette
        self.mode = mode

    def __repr__(self):
        return '<Block mode=%d>' % self.mode

    def calc_error(self, img):
        """Return the error of this block compared to this Image.

        The result is returned as the sum of the squared differences.

        """
        result = 0.0

        # Iterate over each pixel, calculating the error.
        pixels = self.pixmap
        reference = img.getdata()
        palette = self.make_blend_palette()
        for i in range(16):
            value = pixels & 3
            pixels = pixels >> 2
            this_pixel = palette[value]
            ref_pixel = reference[i]
            ssd = sum_squared(calc_error(this_pixel, ref_pixel))
            result += ssd

        return result

    def copy_pixels(self, img):
        """Copy the pixels from this image into our pixmap."""
        data = img.getdata()
        for i in range(15, -1, -1):
            pixel = data[i]
            self.pixmap |= pixel
            if i != 0:
                self.pixmap = self.pixmap << 2

    def make_blend_palette(self):
        """Return this block's palette, plus the mode-dependent blends."""
        if self.mode == 1:
            # Mode 1: 50/50 blend
            blend = tuple(avg(self.palette[0], self.palette[1]))
            pal = self.palette + [blend]
        elif self.mode == 2:
            # Mode 2: no blends
            pal = self.palette
        elif self.mode == 3:
            # Mode 3: 3/8 and 5/8 blends
            blend1 = tuple(avg(self.palette[0], self.palette[1], 5, 3, 8))
            blend2 = tuple(avg(self.palette[0], self.palette[1], 3, 5, 8))
            pal = self.palette + [blend1, blend2]
        return pal

    def make_img(self):
        """Return an Image of how the block will be displayed."""
        # Generate a palette using different blends depending on the mode.
        result = Image.new("P", (4, 4))
        blend_pal = self.make_blend_palette()
        blend_pal = pal_5to8(blend_pal)
        put_palette(result, blend_pal)

        # Copy our pixmap into the image.
        pixels = self.pixmap
        data = []
        for i in range(0, 16):
            data.append(pixels & 3)
            pixels = pixels >> 2
        result.putdata(data)
        return result

    def remap(self, indices):
        """Remap the pixels to different values.

        indices describes the remap as follows:
        indices[0] = 2          change pixels valued 0 to 2
        indices[1] = 1          don't change pixels valued 1
        etc.

        """
        # Put in rules for the blended pixels if they are missing.
        if len(indices) == 2:
            indices += [3, 4]
        elif len(indices) == 3:
            indices += [4]

        # Remap the pixels.
        old = self.pixmap
        self.pixmap = 0
        for i in range(16):
            pixel = indices[old & 3]
            old = old >> 2
            self.pixmap |= (pixel << 2*i)


#
# Block creators
#

def make_mode1_block(block, palette=None):
    """Return a mode 1 block (2-color, with 50/50 blend)."""
    # Quantize to 2 sorted colors if no palette was specified.
    if palette == None:
        q2block = block.quantize(colors=2)
        pal2 = get_palette(q2block, 2)
        pal2.sort()
    else:
        pal2 = palette[0:2]

    # Add a 50/50 blend.
    new_color = avg(pal2[0], pal2[1])
    pal3 = pal2 + [new_color]

    # Quantize to the previous colors and the blend.
    pal_img = Image.new("P", (1, 1))
    put_palette(pal_img, pal3)
    q3block = block.quantize(palette=pal_img)

    # Create the block from the quantized image.
    result = Block(q3block, pal2, 1)
    return result

def make_mode3_block(block, palette=None):
    """Return a mode 3 block (2-color, with 3/8 and 5/8 blends)."""
    # Quantize to 2 sorted colors if no palette was specified.
    if palette == None:
        q2block = block.quantize(colors=2)
        pal2 = get_palette(q2block, 2)
        pal2.sort()
    else:
        pal2 = palette[0:2]

    # Add the blends.
    blend1 = avg(pal2[0], pal2[1], 3, 5, 8)
    blend2 = avg(pal2[0], pal2[1], 5, 3, 8)
    pal4 = pal2 + [blend1, blend2]

    # Quantize to the previous colors and the blend.
    pal_img = Image.new("P", (1, 1))
    put_palette(pal_img, pal4)
    q4block = block.quantize(palette=pal_img)

    # Create the block from the quantized image.
    result = Block(q4block, pal2, 3)
    return result


def make_mode2_block(block, palette=None):
    """Return a mode 2 (4-color) block."""
    if palette == None:
        # Quantize to 4 colors and sort the palette.
        qblock = block.quantize(colors=4)
        pal = get_palette(qblock, 4)
        spal = sorted(pal)
        indices = [spal.index(val) for val in pal]
        remap_image(qblock, indices)
    else:
        # Quantize to the given palette.
        pal_img = Image.new("P", (1, 1))
        put_palette(pal_img, palette)
        qblock = block.quantize(palette=pal_img)
        spal = palette

    # Create the block from the quantized image.
    result = Block(qblock, spal, 2)
    return result

def make_optimal_block(img5, x, y, modes=BLOCK_MODES,
                       threshold=ERROR_THRESHOLD, palette=None):
    """Return an Image representing the best 4x4 block.

    Inputs:
    img5 - The crushed/dithered image (5 bits/channel)
    x, y - The start of the block
    modes - A list of the block types to try.
    threshold - The maximum sum-of-squared-differences that a 2-color palette
                can have before a 4-color palette is tried.
    palette - Optionally, use only these colors.

    """
    # Crop the relevant part of the images.
    block5 = img5.crop((x, y, x+4, y+4))

    # First, try making 2-color palettes.
    if (1 in modes) or (3 in modes):
        m1block = make_mode1_block(block5, palette)
        m3block = make_mode3_block(block5, palette)

        # Calculate the error for each block.
        m1error = m1block.calc_error(block5)
        m3error = m3block.calc_error(block5)

        # Pick the block with the lowest error.
        if m1error <= m3error:
            best_block = m1block
            best_error = m1error
        else:
            best_block = m3block
            best_error = m3error
    else:
        best_error = INF

    # If the error is still too high, try a 4-color palette.
    # Don't do this if we were provided a 2-color palette to start with.
    make_mode2 = (2 in modes) and ((palette is None) or (len(palette) == 4))
    if (best_error > threshold) and make_mode2:
        m2block = make_mode2_block(block5, palette)
        m2error = m2block.calc_error(block5)
        if m2error < best_error:
            best_block = m2block
            best_error = m2error

    return best_block


#
# Image block processing
#

def _block_image_worker(q_jobs, q_results, args):
    """A worker process spawned by block_image()."""
    raw_img5, size, modes, threshold = args
    img5 = Image.fromstring('RGB', size, raw_img5)
    width = size[0]

    # Do each row of blocks from the jobs queue until we get the DONE message.
    while True:
        y = q_jobs.get()
        if y == 'DONE':
            break
        row = []
        for x in range(0, width, 4):
            block = make_optimal_block(img5, x, y, modes, threshold)
            row.append(block)
        q_results.put((y, row))

def block_image(img5, modes=BLOCK_MODES, threshold=ERROR_THRESHOLD):
    """Return a list of optimal blocks made from this image."""
    # Prepare the result array and consumer.
    width, height = img5.size
    bwidth = width/4
    blocks = [None] * (width * height / 16)
    def consume(result):
        y, row = result
        by = y/4
        blocks[(by*bwidth):(by*bwidth + bwidth)] = row

    # Create and start the worker processes.
    raw_img5 = img5.tostring()
    args = (raw_img5, img5.size, modes, threshold)
    pool = WorkerPool(_block_image_worker, args, consume)
    pool.start()

    # Send the list of rows to the workers.
    for y in range(0, height, 4):
        pool.submit(y)

    # Wait for the rest of the results to come in.
    pool.finish()
    return blocks

def unblock_image(blocks, width, height):
    """Return the image created from these blocks.  For testing purposes."""
    img = Image.new("RGB", (width, height))
    i = 0
    for y in range(0, height, 4):
        for x in range(0, width, 4):
            b = blocks[i].make_img()
            img.paste(b, (x, y))
            i += 1
    return img


#
# Image palette processing
#

def remove_duplicate_palettes(blocks):
    """Remove the duplicate palettes from this list of blocks.

    This process is lossless.
    Return a list of unique palettes, and modify the blocks to use it.

    """
    # Load all the unique palettes into all_pals, unreferencing the duplicates
    # from the blocks they were in.  The blocks will be relinked to unique
    # palettes.
    all_pals = {}
    blocks_using = collections.defaultdict(list)
    for b in blocks:
        pal_hash = hash_palette(b.palette)
        same_palette = all_pals.get(pal_hash)
        if same_palette:
            b.palette = same_palette
        else:
            all_pals[pal_hash] = b.palette
        blocks_using[pal_hash].append(b)

    return all_pals.values()

def reduce_palettes(blocks, palettes, palette_size):
    """Reduce the amount of palettes to fit a size in bytes.
    
    The blocks list will be modified to use the new palettes.
    The palettes list will be modified to only contain the new palettes.

    """
    # Split the list of palettes into 2-color and 4-color palettes.
    pal2 = [p for p in palettes if len(p) == 2]
    pal4 = [p for p in palettes if len(p) == 4]
    new_pal2_len, new_pal4_len = calc_pal_size(len(pal2), len(pal4),
                                               palette_size)

    # Use k-means to make some new palettes.  Create a dict mapping the old
    # palettes to the new ones.
    oldhash_to_newpal = dict()
    newpal_all = []
    for pal, new_size in ((pal2, new_pal2_len), (pal4, new_pal4_len)):
        if new_size == 0:
            continue

        # Run k-means on a 6- or 12-dimensional color space, for a 2-color or
        # 4-color palette, respectively.  TODO: Can we do this in parallel?
        entries_per_palette = 3 * len(pal[0])
        flat_pal = numpy.array(pal).reshape(len(pal), entries_per_palette)
        whitened_pal = vq.whiten(flat_pal)
        centroids, indices = vq.kmeans2(whitened_pal, new_size)

        # Transform the centroids back to the original color space.
        std_dev = numpy.std(flat_pal, axis=0)
        centroids = numpy.array(centroids * std_dev, dtype=int)

        # Convert the centroids into palettes, clamping every value to [0,31].
        new_pal = []
        for pal_entry in centroids:
            new_pal_entry = []
            for i in range(0, len(pal_entry), 3):
                r = clamp5(pal_entry[i])
                g = clamp5(pal_entry[i+1])
                b = clamp5(pal_entry[i+2])
                new_pal_entry.append((r, g, b))
            new_pal.append(new_pal_entry)
            newpal_all.append(new_pal_entry)

        # Map the old palette hashes to the new palette entries.
        for old_idx, new_idx in enumerate(indices):
            old_hash = hash_palette(pal[old_idx])
            new_palette_entry = new_pal[new_idx]
            oldhash_to_newpal[old_hash] = new_palette_entry

    # The new palettes are now computed.  Change all the blocks to use them.
    for b in blocks:
        old_hash = hash_palette(b.palette)
        new_palette = oldhash_to_newpal[old_hash]
        b.palette = new_palette

    # Replace all the old palettes with the new ones.
    del palettes[:]
    palettes.extend(newpal_all)


# Originally I wanted to correct all the blocks that aren't good enough.  It's
# just too slow to brute force search the entire palette, even with 4 cores (in
# 2009).  Perhaps with a nice 12-dimensional data structure to search for
# "nearest neighbor" palettes, it would work.
#
# Right now, I am using this only to fix the worst of the worst blocks.
def _reblock_image_worker(q_jobs, q_results, args):
    """Worker function for reblock_image()."""
    # Get the arguments.
    raw_img5, size, blocks, palettes, threshold, correction_threshold = args
    img5 = Image.fromstring('RGB', size, raw_img5)

    # Brute force search each palette until an acceptable error is found.
    while True:
        next = q_jobs.get()
        if next == 'DONE':
            break

        x, y, block_idx, error = next
        curr_block = blocks[block_idx]
        orig_block = img5.crop((x, y, x+4, y+4))
        best_match = curr_block
        best_error = error
        for p in palettes:
            new_block = make_optimal_block(img5, x, y, palette=p)
            new_error = new_block.calc_error(orig_block)
            if new_error < best_error:
                # Better than we had.  Remember this one.
                best_match = new_block
                best_error = new_error
            if new_error <= correction_threshold:
                # In fact, it's good enough.  Stop the search!
                break
    
        # Replace the block with the better one.
        q_results.put((block_idx, best_match))

def reblock_image(img5, blocks, palettes, threshold=EGREGIOUS_THRESHOLD,
    correction_threshold=CORRECTION_THRESHOLD):
    """Redo all blocks that have too high an error.

    Parameters:
    img5 - The 5 bit/channel source image.
    blocks - The blocked version of the image.
    palettes - The list of palettes used by the blocks.
    threshold - The maximum sum of squared differences that is acceptible.
    correction_threshold - The maximum SSD to correct a block to.

    """
    # What to do with the result we get from a worker process.
    def consume(result):
        block_idx, best_match = result
        blocks[block_idx] = best_match

    # Create and start the worker processes.
    raw_img5 = img5.tostring()
    args = (raw_img5, img5.size, blocks, palettes, threshold,
            correction_threshold)
    pool = WorkerPool(_reblock_image_worker, args, consume)
    pool.start()

    # Search for bad blocks to send to the workers.
    width, height = img5.size
    block_idx = 0
    for y in range(0, height, 4):
        for x in range(0, width, 4):
            # Only redo blocks that are past the error threshold.
            orig_block = img5.crop((x, y, x+4, y+4))
            error = blocks[block_idx].calc_error(orig_block)
            if error > threshold:
                pool.submit((x, y, block_idx, error))

            # Continue to the next block.
            block_idx += 1

    # Wait for the rest of the results to come in.
    pool.finish()


#
# Image depth crusher
#

def eight2five_rgb(src):
    """Convert an 8-bit/channel RGB image to a 5-bit/channel RGB image.

    Return a new, non-normalized image (i.e. the image has 5-bit values in
    8-bit fields).

    This function accepts and returns width x height x 3 numpy arrays.

    """
    # Make a copy of src in floating point.
    src = src.astype(float)
    dest = numpy.ndarray(shape=src.shape, dtype=int)

    height, width = src.shape[0:2]
    for y in range(height):
        for x in range(width):
            dest[y,x] = (src[y,x] / 255.0 * 31.0).round()
            dither(src, dest, x, y)

    return dest

DITHER_COEF_1 = 7.0 / 16.0 * 255.0
DITHER_COEF_2 = 3.0 / 16.0 * 255.0
DITHER_COEF_3 = 5.0 / 16.0 * 255.0
DITHER_COEF_4 = 1.0 / 16.0 * 255.0
def dither(src, dest, x, y):
    """Apply floyd-steinberg dithering to the dest image at (x, y)."""
    error = calc_error_arr58(src[y,x], dest[y,x])
    height, width = src.shape[0:2]
    if x != (width - 1):
        src[y, x+1] += error * DITHER_COEF_1
    if y != (height - 1):
        if x != 0:
            src[y+1, x-1] += error * DITHER_COEF_2
        src[y+1, x] += error * DITHER_COEF_3
        if x != (width - 1):
            src[y+1, x+1] += error * DITHER_COEF_4


#
# PIL <-> NumPy conversion
#

def img2arr(img):
    """Convert an RGB PIL Image to a numpy array.
    
    The result will be a (height,width,3) array.

    """
    width, height = img.size
    result = numpy.ndarray(shape=(height, width, 3), dtype=int)
    data = img.load()
    for y in range(height):
        for x in range(width):
            result[y,x] = data[x,y]
    return result

def arr2img(arr):
    """Convert a numpy array to an RGB PIL Image."""
    height, width = arr.shape[0:2]
    result = Image.new("RGB", (width, height))
    data = result.load()
    src = arr.astype(int)
    for y in range(height):
        for x in range(width):
            data[x,y] = tuple(src[y,x])
    return result


#
# Output formatting
#

def nds_format(blocks, palettes):
    """Convert these blocks and palettes into an NDS formatted texture.

    Returns:
    pixmap_data - The pixel data binaries
    index_data - The palette index binaries
    palette_data - The palette binaries
    
    """
    # Convert palettes to binary, while making a map from hashes to offsets.
    pal_offsets = {}
    all_pals = []
    idx = 0
    for p in palettes:
        # Update the offset map.
        h = hash_palette(p)
        pal_offsets[h] = idx

        # Update the palette list.
        bin = binify_palette(p)
        all_pals.append(bin)
        idx += len(bin)

    # Create the final palette binary.
    palette_data = ''.join(all_pals)
    if len(palette_data) > 0x10000:
        raise ValueError, "Too many palettes.  Must fit in 64 KiB."

    # Build the pixmap out of 32-bit words.
    pixmap_chunks = []
    index_chunks = []
    for b in blocks:
        # Pack the pixmaps.
        pixmap_chunks.append(struct.pack("<I", b.pixmap))
        
        # Pack the index/mode data.
        pal_hash = hash_palette(b.palette)
        value = (pal_offsets[pal_hash] >> 2) | (b.mode << 14)
        index_chunks.append(struct.pack("<H", value))

    # Create the rest of the binaries.
    pixmap_data = ''.join(pixmap_chunks)
    index_data = ''.join(index_chunks)

    return (pixmap_data, index_data, palette_data)

def txz_format(pixmap_data, index_data, palette_data, width, height):
    """Return a binary blob of txz formatted data.

    Pass in the output from nds_format() and the image size.

    """
    # Create the header.
    MAGIC = 0x7465787A
    pal_size = len(palette_data) >> 2
    header = struct.pack("<I HH H", MAGIC, width, height, pal_size)

    # Create the compressed body.
    body = ''.join((pixmap_data, index_data, palette_data))
    zbody = zlib.compress(body, 9)
    
    # Return the combined result.
    return ''.join((header, zbody))


#
# Utilities
#

def get_palette(img, num_colors):
    """Return this Image's palette, using exactly num_colors."""
    # The most ridiculous way to access a palette I've ever seen, from:
    # http://www.pythonware.com/library/pil/handbook/imagepalette.htm
    pal_img = img.resize((num_colors, 1))
    pal_img.putdata(range(num_colors))
    pal_img = pal_img.convert("RGB")
    return list(pal_img.getdata())

def put_palette(img, palette):
    """Replace this Image's palette with a list of color tuples."""
    # Start with a flat version of the palette.
    new_pal = flatten_palette(palette)

    # Fill the empty slots with color 0.
    while len(new_pal) < 256*3:
        new_pal.extend(palette[0])

    img.putpalette(new_pal)


def flatten_palette(p):
    """Return a flat list of palette entries, instead of a list of tuples."""
    flat = []
    for entry in p:
        flat.extend(entry)
    return flat


def binify_palette(p):
    """Convert this palette to binary NDS format (15 bit)."""
    bin_colors = []
    for entry in p:
        r, g, b = entry
        c = (b << 10) | (g << 5) | r
        bin_colors.append(struct.pack("<H", c))
    return ''.join(bin_colors)


def palette_size(p):
    """Return the size in bytes of this palette."""
    return sum([len(e) for e in p]) * 2


def hash_palette(p):
    """Return a hash for a palette."""
    return hash(tuple(flatten_palette(p)))


def remap_image(img, indices):
    """Remap the pixels to different values.

    indices describes the remap as follows:
    indices[0] = 2          change pixels valued 0 to 2
    indices[1] = 1          don't change pixels valued 1
    etc.

    """
    # Remap the pixels.
    data = img.load()
    width, height = img.size
    for y in range(height):
        for x in range(width):
            pixel = data[x,y]
            data[x,y] = indices[pixel]


def pal_5to8(pal):
    """Return an 8 bit/channel version of this 5 bit/channel palette."""
    new_pal = []
    for entry in pal:
        new_entry = []
        for channel in entry:
            new_entry.append(channel * 255 / 31)
        new_pal.append(new_entry)
    return new_pal

def clamp5(x):
    """Return the value clamped to 5 bits [0,31]."""
    return max(0, min(31, x))

def calc_pal_size(len_pal2, len_pal4, target_size, minimum_len=8):
    """Calculate the palette sizes to fit in this amount of bytes.

    Parameters:
    len_pal2 - The number of entries in pal2.
    len_pal4 - The number of entries in pal4.
    target_size - The size they should fit in, in bytes.
    minimum_len - The smallest amount of entries in either palette.

    Returns:
    new_pal2_len - The new number of entries in pal2.
    new_pal4_len - The new number of entries in pal4.

    """
    # Convert everything to bytes.
    size_pal2 = len_pal2 * 4
    size_pal4 = len_pal4 * 8
    size_total = size_pal2 + size_pal4
    min_size_2 = minimum_len * 4
    min_size_4 = minimum_len * 8

    # Make a proportional amount of 2-color and 4-color palettes.
    new_size_pal2 = size_pal2 * target_size / size_total

    # Adjust that number so that the mininmum number of entries are created.
    adj_size_pal2 = min(size_pal2, max(min_size_2, new_size_pal2))

    # Make pal4 from what's left, also adjusting to the minimum.
    new_size_pal4 = target_size - adj_size_pal2
    adj_size_pal4 = min(size_pal4, max(min_size_4, new_size_pal4))

    # If pal4 was adjusted, take some off of pal2 to get back to the total.
    diff = adj_size_pal4 - new_size_pal4
    adj_size_pal2 -= diff

    new_pal2_len = adj_size_pal2 / 4
    new_pal4_len = adj_size_pal4 / 8
    return (new_pal2_len, new_pal4_len)


def calc_error(pixel1, pixel2):
    """Return an array of (R, G, B) error between two RGB5 pixels.
    
    pixel1 - 5 bit/channel RGB in a python list.
    pixel2 - 5 bit/channel RGB in a python list.

    """
    result = []
    for i in range(3):
        error = float(pixel1[i] - pixel2[i]) / 31.0
        result.append(error)
    return result

def calc_error_arr58(rgb8, rgb5):
    """Return an array of (R, G, B) error between an RGB8 and RGB5 pixel.
    
    rgb8 - 8 bit/channel RGB in a numpy array.
    rgb5 - 5 bit/channel RGB in a numpy array.
    
    """
    return (rgb8 / 255.0) - (rgb5 / 31.0)

def sum_squared(l):
    """Return the sum of the squared items in l."""
    return sum([n*n for n in l])

def avg(l1, l2, numerator1=1, numerator2=1, denominator=2):
    """Return a list, each item being the average of an item in l1 and l2."""
    result = []
    for i in range(len(l1)):
        result.append((numerator1*l1[i] + numerator2*l2[i]) / denominator)
    return result

def write_data(fname, data):
    """Write this binary data to this filename."""
    f = open(fname, "wb")
    f.write(data)
    f.close()


#
# Multiprocessing worker pool
#

class WorkerPool:
    def __init__(self, target, args, consumer):
        """Create a pool to run some tasks.

        Inputs:
        target - A *non-nested* function to run in each worker process.  Its
                 signature is result = target(q_jobs, q_results, args).  When
                 target receives the string 'DONE' in q_jobs, it should return.
        args - An argument object to pass to target.
        consumer - The function to run (nested is OK) when a result is ready.
                   Its signature is consumer(result).

        """
        self.target = target
        self.args = args
        self.consumer = consumer
        self.processes = []
        self.q_jobs = multiprocessing.Queue()
        self.q_results = multiprocessing.Queue()
        self.num_jobs_finished = 0
        self.num_jobs_submitted = 0

    def start(self):
        """Start running all worker processes."""
        args = (self.q_jobs, self.q_results, self.args)
        for i in range(multiprocessing.cpu_count()):
            p = multiprocessing.Process(target=self.target, args=args)
            self.processes.append(p)
            p.start()

    def finish(self):
        """Wait for all jobs to finish, and terminate workers."""
        # Get all the remaining results.
        while self.num_jobs_finished != self.num_jobs_submitted:
            self._get_and_consume_result()

        # Terminate the workers.
        [self.q_jobs.put('DONE') for p in self.processes]
        [p.join() for p in self.processes]

    def submit(self, job):
        """Submit this job object by placing it in the q_jobs Queue."""
        # If the job queue is full, wait for a result to come in and try again.
        put_complete = False
        while not put_complete:
            try:
                self.q_jobs.put(job, False)
                put_complete = True
            except Queue.Full:
                self._get_and_consume_result()
        self.num_jobs_submitted += 1

    def _get_and_consume_result(self):
        result = self.q_results.get()
        self.num_jobs_finished += 1
        self.consumer(result)


#
# Commandline interface
#

def print_usage(msg=None):
    """Print the usage information."""
    print __doc__
    print \
"""USAGE: dstexcomp.py <input image> [options]

Long                    Short       Description
OUTPUT OPTIONS
--binary-out=<base>     -b <base>   Write raw <base>_{tex,idx,pal}.bin files
--preview-out=<file>    -p <file>   Write a preview image file
--compressed-out=<file> -z <file>   Write a compressed .txz file

COMPRESSION OPTIONS
--palette-size=<size>   -s <size>   The palette size in bytes (def: %d)
--block-modes=<list>    -m <list>   Use these comma-separated modes (def: %s)
--egregious-threshold=  -e <err>    Minimum "egregiously bad" error (def: %s)
--correction-threshold= -c <err>    New minimum error for -e blocks (def: %s)

OTHER OPTIONS
--help                  -h          Print this help message
--version               -v          Print version info and exit
--quiet                 -q          Do not print any informational messages"""\
    % (PALETTE_SIZE,
       ','.join([str(n) for n in BLOCK_MODES]),
       str(EGREGIOUS_THRESHOLD),
       str(CORRECTION_THRESHOLD))

    if msg:
        print
        print msg


def parse_commandline():
    """Return a dict of options parsed from the commandline."""
    # The default options dict:
    result = {
        'input': None,
        'binary_out': None,
        'preview_out': None,
        'compressed_out': None,
        'palette_size': PALETTE_SIZE,
        'block_modes': BLOCK_MODES,
        'egregious_threshold': EGREGIOUS_THRESHOLD,
        'correction_threshold': CORRECTION_THRESHOLD,
        'quiet': False,
    }

    try:
        opts, args = getopt.gnu_getopt(sys.argv[1:], "b:p:z:s:m:e:c:hqv",
                     ["binary-out=", "preview-out=", "compressed-out=",
                      "palette-size=", "block-modes=", "egregious-threshold=",
                      "correction-threshold=", "help", "quiet", "version"])
    except getopt.GetoptError, err:
        print_usage(str(err))
        sys.exit()

    # Fill in the fields.
    for o, a in opts:
        if o in ('-b', '--binary-out'):
            result['binary_out'] = a
        elif o in ('-p', '--preview-out'):
            result['preview_out'] = a
        elif o in ('-z', '--compressed-out'):
            result['compressed_out'] = a
        elif o in ('-s', '--palette-size'):
            result['palette_size'] = int(a)
        elif o in ('-m', '--block-modes'):
            modes = [int(n) for n in a.split(',')]
            result['block_modes'] = modes
        elif o in ('-e', '--egregious-threshold'):
            result['egregious_threshold'] = float(a)
        elif o in ('-c', '--correction-threshold'):
            result['correction_threshold'] = float(a)
        elif o in ('-q', '--quiet'):
            result['quiet'] = True
        elif o in ('-h', '--help'):
            print_usage()
            sys.exit()
        elif o in ('-v', '--version'):
            print "dstexcomp", VERSION
            sys.exit()

    if len(args) == 0:
        print_usage("No input file specified.")
        sys.exit()
    if len(args) > 1:
        print_usage("Too many arguments given.  Only specify one input image.")
        sys.exit()
    result['input'] = args[0]

    if (result['binary_out'] is None) and (result['preview_out'] is None) and \
       (result['compressed_out'] is None):
        print_usage("No output file specified.")
        sys.exit()

    return result


class InfoPrinter:
    def __init__(self, quiet):
        self.quiet = quiet

    def out(self, s):
        if not self.quiet:
            print s


#
# Main function
#


def main():
    # Use psyco if available.
    try:
        import psyco
        psyco.full()
    except ImportError:
        pass

    # Parse the commandline.
    opts = parse_commandline()
    info = InfoPrinter(opts['quiet'])

    # Load the source image, both as an Image and numpy.array.
    info.out("Loading the source image...")
    src8 = Image.open(opts['input'])
    width, height = src8.size
    valid_dimensions = (8, 16, 32, 64, 128, 256, 512, 1024)
    if (width not in valid_dimensions) or (height not in valid_dimensions):
        warnings.warn("Image size is not valid (power of two, <= 1024)")
    arr8 = img2arr(src8)

    # Crush the depth down to 5 bits/channel, and convert back to an Image.
    info.out("Reducing depth to 5 bits/channel...")
    arr5 = eight2five_rgb(arr8)
    src5 = arr2img(arr5)

    # Create the best possible compressed image.
    info.out("Compressing the image...")
    blocks = block_image(src5, opts['block_modes'])
    palettes = remove_duplicate_palettes(blocks)

    # If the palette is too large, reduce it.
    size = opts['palette_size']
    if palette_size(palettes) > size:
        info.out("Reducing palette to %d bytes..." % size)
        # K-means is going to raise warnings no matter what, so ignore them.
        warnings.simplefilter('ignore', UserWarning)
        reduce_palettes(blocks, palettes, size)
        warnings.simplefilter('default', UserWarning)
    else:
        info.out("No palette reduction necessary.")

    # Correct egregiously bad blocks.
    info.out("Correcting egregiously bad blocks...")
    threshold = opts['egregious_threshold']
    correction = opts['correction_threshold']
    reblock_image(src5, blocks, palettes, threshold, correction)

    # Write the output files.
    pixmap_data, index_data, palette_data = nds_format(blocks, palettes)
    if opts['binary_out']:
        info.out("Writing binary output...")
        base = opts['binary_out']
        write_data(base + '_tex.bin', pixmap_data)
        write_data(base + '_idx.bin', index_data)
        write_data(base + '_pal.bin', palette_data)
    if opts['compressed_out']:
        info.out("Writing compressed output...")
        data = txz_format(pixmap_data, index_data, palette_data, width, height)
        write_data(opts['compressed_out'], data)
    if opts['preview_out']:
        info.out("Reconstructing and writing preview output...")
        preview = unblock_image(blocks, width, height)
        preview.save(opts['preview_out'])

    info.out("Exiting normally.")

#
# Entry point
#

if __name__ == '__main__':
    main()
